/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Sun Feb 24 16:10:59 1991 by burrows        */
#include <stdio.h>
#include <ctype.h>
#include <signal.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <errno.h>

#include "indexserver.h"

#ifndef ACTIVEFILE
#define ACTIVEFILE "/usr/lib/news/active"
#endif

#define DELAY	0
#define QUANTUM	60

static int printmsg ();
static int printhdr ();
static int printsummary ();
static void updatei ();
static void update ();
static void update0 ();
static void update1 ();
static void intr ();

static int alarmdisable = 1;
static int alarmpending = 0;
int interrupt = 0;

static char groupsfile[128];
static char deletefile[128];
static char messagesfile[128];
static char concordancefile[128];
static char summaryfile[128];
static char datesfile[128];
static char concordancerecentfile[128];
static char summaryrecentfile[128];
static char datesrecentfile[128];
static char activefile[128];

static struct index2 index2;
static struct groups groups;

static char *prog = "makenewsindex";
static char *indexdir = 0;

static FILE *logfile;

static void coredump (sig)
int sig;
{
	abort ();
}

static void usage ()
{
	(void) fprintf (stderr,
		"usage: indexserver [-f logfile] indexdir [prog]\n");
	exit (2);
}

main (argc, argv)
int argc;
char *argv[];
{
	char commandbuf[5*1024];
	char cmdstr[100];
	register struct msgset *msgset;
	char *inbuf = malloc (BUFSIZ);
	char *outbuf = malloc (BUFSIZ);
	int argn;
	char *arg;
	char *lfile = 0;

#ifdef SETLOCALE
        (void) setlocale (0, "");
#endif

	for (argn = 1; argn < argc && argv[argn][0] == '-'; argn++) {
		for (arg = &argv[argn][1]; *arg; arg++) {
			switch (*arg)
			{
			case 'f':
				if (argn + 1 >= argc)
					usage ();
				lfile = argv[++argn];
				break;
			default: usage ();
			}
		}
	}

	if (argn < argc) 
		indexdir = argv[argn++];
	if (argn < argc) 
		prog = argv[argn++];
	if (argn < argc || indexdir == 0)
		usage ();

	if (strlen (indexdir) + 30 > 1024) {
		(void) fprintf (stderr, "indexserver: argument too long\n");
		exit (2);
	}

	if (lfile != 0) {
		if ((logfile = fopen (lfile, "a")) == 0) {
			perror (lfile);
			exit (2);
		}
	}

	setbuf (stdin, inbuf);
	setbuf (stdout, outbuf);
	tableinitialize ();

	(void) sprintf (groupsfile, "%s/current/groups", indexdir);
	(void) sprintf (deletefile, "%s/current/delete", indexdir);
	(void) sprintf (messagesfile, "%s/current/messages", indexdir);
	(void) sprintf (summaryfile, "%s/current/summary", indexdir);
	(void) sprintf (datesfile, "%s/current/dates", indexdir);
	(void) sprintf (concordancefile, "%s/current/index", indexdir);
	(void) sprintf (summaryrecentfile, "%s/current/summary.recent",
		indexdir);
	(void) sprintf (datesrecentfile, "%s/current/dates.recent", indexdir);
	(void) sprintf (concordancerecentfile, "%s/current/index.recent",
		indexdir);
	(void) strcpy (activefile, ACTIVEFILE);

	update (0);

	(void) signal (SIGCHLD, updatei);
	(void) signal (SIGALRM, updatei);
	(void) signal (SIGINT, intr);
	(void) signal (SIGSEGV, coredump);
	(void) signal (SIGQUIT, coredump);
	(void) signal (SIGBUS, coredump);
	(void) alarm (QUANTUM);

	alarmdisable = 0;
	while (fgets (commandbuf, sizeof (commandbuf), stdin)) {
		char *cmd, *rest, *p;
		int counttoo = 0;
		int error = 0;

		alarmdisable = 1;

		if (logfile != 0)
			(void) fprintf (logfile, "%s", commandbuf);

		if ((p = index (commandbuf, '\n')) != 0 && p[1] == '\0')
			*p = '\0';

		p = commandbuf;
		while (*p != '\0' && isspace (*p))
			p++;
		cmd = cmdstr;
		while (*p != '\0' && isalnum (*p)) {
			if (cmd < &cmdstr[sizeof (cmdstr) - 1])
				*cmd++ = *p;
			p++;
		}
		*cmd = '\0';
		cmd = cmdstr;
		while (*p != '\0' && isspace (*p))
			p++;
		if (*p == '-') {
			p++;
			while (isalnum (*p)) {
				switch (*p)
				{
				case 'c':	counttoo = 1; break;
				default:	/* bad flag */; break;
				}
				p++;
			}
			while (*p != '\0' && isspace (*p))
				p++;
		}
		rest = p;
		while (*p != '\0' && !isspace (*p))
			p++;

		if (strlen (cmd) == 0) {
			/* nothing */;
		} else if (strncmp (cmd, "summary", strlen (cmd)) == 0) {
			msgset = getmsgset (&index2, &groups, rest, &error);
			if (!error) {
				if (counttoo)
					(void) printf ("%ld\n",
						countmsgset (msgset));
				(void) applymsgset (printsummary, &index2,
					&groups, msgset);
			}
			freemsgset (msgset);
		} else if (strncmp (cmd, "headers", strlen (cmd)) == 0) {
			msgset = getmsgset (&index2, &groups, rest, &error);
			if (!error) {
				if (counttoo)
					(void) printf ("%ld\n",
						countmsgset (msgset));
				(void) applymsgset (printhdr, &index2,
					&groups, msgset);
			}
			freemsgset (msgset);
		} else if (strncmp (cmd, "count", strlen (cmd)) == 0) {
			msgset = getmsgset (&index2, &groups, rest, &error);
			if (!error)
				(void) printf ("%ld\n", countmsgset (msgset));
			freemsgset (msgset);
		} else if (strncmp (cmd, "files", strlen (cmd)) == 0) {
			msgset = getmsgset (&index2, &groups, rest, &error);
			if (!error) {
				if (counttoo)
					(void) printf ("%ld\n",
						countmsgset (msgset));
				(void) applymsgset (printmsg, &index2,
					&groups, msgset);
			}
			freemsgset (msgset);
		} else {
			(void) printf ("?%s: unknown command\n", cmd);
		}

		(void) fputc ('\0', stdout);
		interrupt = 0;
		(void) fflush (stdout);

		if (alarmpending)
			update (0);
		alarmdisable = 0;
	}

	exit (0);
}

static int printmsg (index2, groups, msg)
register struct index2 *index2;
register struct groups *groups;
register long msg;
{
	register struct message *m;
	register long mindex = msg * sizeof (struct message);
	register long n;

	if (mindex < (n = index2->indices[0].summary.messages.ndata)) {
		m = (struct message *)
			(index2->indices[0].summary.messages.data + mindex);
	} else if (mindex < n + index2->indices[1].summary.messages.ndata) {
		m = (struct message *)
			(index2->indices[1].summary.messages.data + mindex - n);
	} else {
		(void) fprintf (stderr,
		    "indexserver: printmsg: message %d not in either index (%d %d %d)\n",
		    msg, mindex, index2->indices[0].summary.messages.ndata,
		    index2->indices[1].summary.messages.ndata);
		abort ();
	}

	(void) printf ("%s/%d\n",
		groups->text.text.data + groups->group[m->group], m->message);
	    
	return (interrupt);
}

static int printhdr (index2, groups, msg)
register struct index2 *index2;
register struct groups *groups;
register long msg;
{
	register struct message *m;
	char datebuf[30];
	register long mindex = msg * sizeof (struct message);
	register long n;
	char *f;

	if (mindex < (n = index2->indices[0].summary.messages.ndata)) {
		m = (struct message *)
			(index2->indices[0].summary.messages.data + mindex);
	} else if (mindex < n + index2->indices[1].summary.messages.ndata) {
		m = (struct message *)
			(index2->indices[1].summary.messages.data + mindex - n);
	} else {
		(void) fprintf (stderr,
		    "indexserver: printhdr: message %d not in either index (%d %d %d)\n",
		    msg, mindex, index2->indices[0].summary.messages.ndata,
		    index2->indices[1].summary.messages.ndata);
		abort ();
	}

	(void) printf ("File: %s/%d\n",
		groups->text.text.data + groups->group[m->group], m->message);
	f = field (index2, msg, FROMFIELD);
	if (*f != '\0') (void) printf ("From: %s\n", f);
	f = field (index2, msg, TOFIELD);
	if (*f != '\0') (void) printf ("To: %s\n", f);
	f = field (index2, msg, CCFIELD);
	if (*f != '\0') (void) printf ("Cc: %s\n", f);
	f = field (index2, msg, SUBJECTFIELD);
	if (*f != '\0') (void) printf ("Subject: %s\n", f);
	(void) printf ("Date: %s\n", 
		unparsedate (strtol (field (index2, msg, DATEFIELD),
			(char **)0, 16), datebuf, 1));
	f = field (index2, msg, NEWSGROUPSFIELD);
	if (*f != '\0') (void) printf ("Newsgroups: %s\n", f);
	f = field (index2, msg, MESSAGEIDFIELD);
	if (*f != '\0') (void) printf ("Message-Id: %s\n", f);
	f = field (index2, msg, INREPLYTOFIELD);
	if (*f != '\0') (void) printf ("In-Reply-To: %s\n", f);
	f = field (index2, msg, REFERENCESFIELD);
	if (*f != '\0') (void) printf ("References: %s\n", f);
	f = field (index2, msg, KEYWORDSFIELD);
	if (*f != '\0') (void) printf ("Keywords: %s\n", f);
	(void) printf ("\n");
	    
	return (interrupt);
}

static int printsummary (index2, groups, msg)
register struct index2 *index2;
register struct groups *groups;
register long msg;
{
	register struct message *m;
	char datebuf[30];
	register long mindex = msg * sizeof (struct message);
	register long n;
	char filebuf[200];

	if (mindex < (n = index2->indices[0].summary.messages.ndata)) {
		m = (struct message *)
			(index2->indices[0].summary.messages.data + mindex);
	} else if (mindex < n + index2->indices[1].summary.messages.ndata) {
		m = (struct message *)
			(index2->indices[1].summary.messages.data + mindex - n);
	} else {
		(void) fprintf (stderr,
		    "indexserver: printsummary: message %d not in either index (%d %d %d)\n",
		    msg, mindex, index2->indices[0].summary.messages.ndata,
		    index2->indices[1].summary.messages.ndata);
		abort ();
	}

	(void) sprintf (filebuf, "%s/%d", 
		groups->text.text.data + groups->group[m->group], m->message);
	(void) printf ("%-19s %-9.9s %s %.41s\n",
		filebuf,
		field (index2, msg, FROMFIELD),
		unparsedate (strtol (field (index2, msg, DATEFIELD),
			(char **)0, 16), datebuf, 0),
		field (index2, msg, SUBJECTFIELD));
	    
	return (interrupt);
}

static void updatei ()
{
	update (1);
}

static void update (i)
int i;
{
	static int pid = -1;
	static long lastrun = 0;
	static long lastsummarytime = 0;
	static long lastsummaryrecenttime = 0;
	static long lastactivetime = 0;
	int dead;
	union wait status;
	struct stat infosummary, infosummaryrecent, infoactive;
	int save;
	long now;
	
	save = alarmdisable;
	alarmdisable = 1;

	if (save && i) {
		alarmpending = 1;
		alarmdisable = save;
		return;
	}
	alarmpending = 0;

	if (logfile != 0)
		(void) fflush (logfile);

	if (pid > 0) {
		do {
			dead = wait3 (&status, WNOHANG, (struct rusage *) 0);

			if (dead == pid) {
				pid = -1;
				lastrun = time ((long *) 0);
			}
		} while (dead > 0 || errno == EINTR);
	}
	if (pid > 0) {
		(void) alarm (QUANTUM);
		alarmdisable = save;
		return;
	}

	if (stat (summaryfile, &infosummary) < 0) {
		perror (summaryfile);
		exit (2);
	}
	if (stat (summaryrecentfile, &infosummaryrecent) < 0) {
		perror (summaryrecentfile);
		exit (2);
	}

	if (infosummary.st_mtime != lastsummarytime) {
		lastsummarytime = infosummary.st_mtime;
		lastsummaryrecenttime = infosummaryrecent.st_mtime;
		update0 ();
	} else if (infosummaryrecent.st_mtime != lastsummaryrecenttime) {
		lastsummaryrecenttime = infosummaryrecent.st_mtime;
		update1 ();
	}

	if (lastrun + DELAY < (now = time ((long *) 0)) &&
	    stat (activefile, &infoactive) == 0 &&
	    lastactivetime != infoactive.st_mtime) {
		int coalesce = 0;

		lastactivetime = infoactive.st_mtime;

		if (infosummaryrecent.st_size > 100000) {
			struct tm *tm = localtime (&now);

			if (tm->tm_hour < 6)
				coalesce = 1;
		}

		pid = vfork ();
		if (pid == -1) {
			perror ("vfork");
			exit (2);
		}

		if (pid == 0) {
			(void) alarm (0);
			if (chdir (indexdir) < 0) {
				perror (prog);
			} else if (coalesce) {
				(void) execlp (prog, prog, "-c", (char *) 0);
				perror (prog);
			} else {
				(void) execlp (prog, prog, (char *) 0);
				perror (prog);
			}
			(void) _exit (2);
		}
	}
	(void) alarm (QUANTUM);
	alarmdisable = save;
}

static void update0 ()
{
	freeindex (&index2.indices[0]);
	freeindex (&index2.indices[1]);
	freegroups (&groups);
	clearvars ();

	freemsgset (index2.allmsgs);
	index2.allmsgs = 0;

	if (readindex (&index2.indices[0], 0L, messagesfile,
			concordancefile, summaryfile, datesfile) < 0)
		exit (2);

	update1 ();
}

static void update1 ()
{
	register int i, n;
	register long *x;
	register FILE *fp;
	long d;

	clearvars ();
	freeindex (&index2.indices[1]);
	freegroups (&groups);

	freemsgset (index2.allmsgs);
	index2.allmsgs = 0;

	if (readindex (&index2.indices[1],
	    index2.indices[0].summary.messages.ndata, messagesfile,
	    concordancerecentfile, summaryrecentfile, datesrecentfile) < 0)
		exit (2);

	if (readgroups (&groups, groupsfile) < 0)
		exit (2);

	n = index2.indices[0].summary.summary.nlines +
	    index2.indices[1].summary.summary.nlines;
	index2.allmsgs = newmsgset ((int) ROUNDUP (n));
        x = index2.allmsgs->bitset;
        i = index2.allmsgs->nbitset;
        i++;
        while (--i != 0)
                *x++ = ~0;
        i = n;
        while (i < index2.allmsgs->nbitset*LONGBITS) {
                index2.allmsgs->bitset[ROUNDDOWN (i)] &= ~(1 << BIT (i));
                i++;
        }

	if ((fp = fopen (deletefile, "r")) == 0) {
		perror (deletefile);
		exit (2);
	}

	x = index2.allmsgs->bitset;
	while (fscanf (fp, "%ld\n", &d) == 1) {
		if (d < 0 || d >= n) {
			(void) fprintf (stderr,
			    "delete on non-existent file %ld n=%d\n", d, n);
		} else {
			x[ROUNDDOWN (d)] &= ~(1 << BIT (d));
		}
	}

	if (!feof (fp) || ferror (fp)) {
		perror (deletefile);
		exit (2);
	}
	(void) fclose (fp);
}

static void intr ()
{
	interrupt = 1;
}

void myabort ()
{
#undef abort
	extern void abort ();

	(void) fprintf (stderr, "indexserver: writing core file\n");
	/* make sure we have permission */
	(void) umask (0);
	(void) creat ("core", 0777);
	(void) chmod ("core", 0777);
	(void) setuid (getuid ());
	abort ();
}
