/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Wed Feb 27 15:22:57 1991 by burrows        */
#include <stdio.h>
#include <strings.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/file.h>
#include <errno.h>
#include <sgtty.h>
#include <signal.h>
#include <setjmp.h>
#include <errno.h>
#include <sys/wait.h>

#ifndef SERVERHOST
#define SERVERHOST "niserver"
#endif

#define INTR -2
#define REGION  (sig_nostop++)
#define ENDREGION       do { if (--sig_nostop <= 0) endregion (); } while (0)
static int sig_interrupt = 0;
static int sig_interrupted = 0;
static int sig_stop = 0;
static int sig_nostop = 0;

extern char *malloc ();
extern char *realloc ();
extern void free ();
extern void bcopy ();
extern void perror ();
extern void exit ();
extern void abort ();
extern char *getenv ();

static int interactive;
static int tenex;
static int erasechar = '\177';

static char *agetline ();
static char *fgetline ();
static void cmdloop ();
static void quit ();
static int gettty ();
static void fixtty ();
static int tenexmsgset ();
static int isescape ();
static int inputstring ();
static char *tgetline ();
static void dotenex ();
static char *getarg ();
static void helpcmd ();
static void quitcmd ();
static void sendcmd ();
static void send ();
static void typecmd ();
static void nullcmd ();
static char *chartos ();
static void sigint ();
static void sigtstp ();
static void endregion ();
static int intread ();
static int intwrite ();
static int intgetc ();
static void xargs ();
static void page ();

struct cmdtable {
	int c;
	char *word;
	void (*action) ();
	int arg;
	int msgset;
	char *help;
};

struct cmdtable topcmds[] = {
    {EOF,   "quit",     quitcmd,    0, 0, (char *) 0},
    {'\004',"quit",     quitcmd,    0, 0, (char *) 0},
    {'\n',  "",         nullcmd,    0, 0, (char *) 0},
    {'q',   "quit",     quitcmd,    0, 0, "quit"},
    {'c',   "count",    sendcmd,    0, 1, "count messages selected"},
    {'s',   "summary",  sendcmd,    0, 1, "show message summaries (one per line)"},
    {'h',   "headers",  sendcmd,    0, 1, "show message headers"},
    {'t',   "type",     typecmd,    0, 1, "type messages selected"},
    {'f',   "files",    sendcmd,    0, 1, "file names of selected messages"},
    {'?',   "?",        helpcmd,    0, 0, "show this information"},
    {0,     (char *) 0, quitcmd,    0, 0, (char *) 0}
};

struct op {
        char tc;
        char *op;
        int expectarg;          /* 1 = string arg, 2 = msgset arg */
	char *help;
};

struct op ops[] = {
        { 'A',  "All",   	0,	"all messages"},
        { 'F',  "From_to_cc",   1,
		"From_to_cc (str) == (from (str) | to (str) | cc (str))"},
        { 'I',  "Id",   	1,
		"Id (str) == (messageid (str) | inreplyto (str))"},
        { 'T',  "To_cc",        1,	"To_cc (str) == (to (str) | cc (str))"},
        { 'a',  "after",        1,	"after date"},
        { 'b',  "before",       1,	"before date"},
        { 'c',  "cc",           1,	"Cc: field"},
        { 'd',  "discussion",   2,	"msgs in same discussion as set"},
        { 'e',  "earliest",     2,	"earliest message in set"},
        { 'f',  "from",         1,	"From: field"},
        { 'g',  "group",        1,	"Newsgroup: field"},
        { 'i',  "inreplyto",    1,	"In-Reply-To: field"},
        { 'k',  "keyword",      1,	"Keyword: field"},
        { 'l',  "latest",       2,	"latest message in set"},
        { 'm',  "messageid",    1,	"Message-Id: field"},
        { 'o',  "on",           1,	"on date (syntax: 12 jan 89)"},
        { 'p',  "pathname",     1,	"e.g src/bugs/17"},
        { 'r',  "references",   1,	"References: field"},
        { 's',  "subject",      1,	"Subject: field"},
        { 't',  "to",           1,	"To: field"},
        { 'w',  "word",         1,	"contains word"},

        /* { 'W',  "Word",         1,	"contains word (case dependent)"}, */
        { 0,   (char *) 0 }
};

struct args {
	int argc;
	int argn;
	char **argv;
	int i;
};

static void usage (prog)
char *prog;
{
	if (prog == 0)
		prog = "ni";

	if (rindex (prog, '/') != 0)
		prog = rindex (prog, '/') + 1;

	(void) fprintf (stderr, "usage: %s [-itn] [-h host] [request]\n", prog);

	quit (2);
}

#define DEFAULTPORT	"2314"

main (argc, argv)
int argc;
char *argv[];
{
	int argn;
	int i;
	struct sockaddr_in addr;
	char *hostname = SERVERHOST;
	char *port = 0;
	struct hostent *hp;
	struct servent *sp;
	char *arg;

#ifdef SETLOCALE
	(void) setlocale (0, "");
#endif

	interactive = -1;
	tenex = 0;

	setlinebuf (stderr);

	for (argn = 1; argn < argc; argn++) {
		arg = argv[argn];
		if (arg[0] == '-') {
			for (i = 1; arg[i] != '\0'; i++) {
				switch (arg[i])
				{
				case 'h':
					if (argn+1 >= argc)
						usage (argv[0]);
					hostname = argv[++argn];
					break;
				case 'n': interactive = 0; break;
				case 'i': interactive = 1; break;
				case 't': tenex = -1; break;
				case 'p':
					if (argn+1 >= argc)
						usage (argv[0]);
					port = argv[++argn];
					break;
				default: usage (argv[0]);
				}
			}
		} else
			break;
	}

	if (interactive < 0 || tenex != 0) {
		if (argn >= argc)
			(void) gettty (0, 2);
	}
	if (interactive < 0)
		interactive = 0;
	if (tenex < 0 || interactive == 0)
		tenex = 0;

	if ((hp = gethostbyname (hostname)) == 0) {
		(void) fprintf (stderr, "?ni: unknown host: %s\n", hostname);
		quit (2);
	}

	addr.sin_family = hp->h_addrtype;
	bcopy (hp->h_addr, (char *) &addr.sin_addr, hp->h_length);
	endhostent ();

	if (port != 0)
		addr.sin_port = htons (atoi (port));
	else if ((sp = getservbyname ("newsindex", "tcp")) == 0)
		addr.sin_port = htons (atoi (DEFAULTPORT));
	else
		addr.sin_port = sp->s_port;
	endservent ();

	if (argn < argc) {
		struct args args;

		args.argc = argc;
		args.argn = argn;
		args.argv = argv;
		args.i = 0;

		cmdloop (agetline, (char *) &args, &addr);
	} else if (tenex) {
		(void) signal (SIGTSTP, sigtstp);
		(void) signal (SIGINT, sigint);
		cmdloop (tgetline, (char *) stdin, &addr);
	} else {
		if (interactive)
			(void) signal (SIGINT, sigint);
		cmdloop (fgetline, (char *) stdin, &addr);
	}

	quit (0);
}

static int makesocket (addr)
struct sockaddr_in *addr;
{
	int fd;
	int one = 1;

	if ((fd = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
		perror ("?ni: socket");
		quit (2);
	}
	if (connect (fd, (struct sockaddr *) addr,
			sizeof (struct sockaddr_in)) < 0) {
		perror ("?ni: connect");
		quit (2);
	}

	if (setsockopt (fd, SOL_SOCKET, SO_KEEPALIVE,
			(char *) &one, sizeof (one)) < 0) {
		perror ("?ni: SO_KEEPALIVE");
	}
	return (fd);
}

static void quit (rc)
int rc;
{
	if (tenex)
		fixtty (0, 0);
	exit (rc);
}

static void cmdloop (func, arg, addr)
char *(*func) ();
char *arg;
struct sockaddr_in *addr;
{
	char *line = 0;
	int fd = -1;
	char cmd[50];
	int i;
	struct cmdtable *cmds = topcmds;

	if (tenex)
		fixtty (0, 1);

	for (;;) {
		if (interactive)
			(void) fprintf (stderr, "<- ");
		
		if (fd < 0)
			fd = makesocket (addr);

		if (line != 0)
			free (line);

		line = (*func) (arg);

		if (line == 0)
			return;

		(void) getarg (line, cmd, sizeof (cmd));

		for (i = 0; cmds[i].word != 0; i++) {
			if (strlen (cmd) == 0) {
				if (cmds[i].word[0] == '\0')
					break;
			} else if (cmd[0] == cmds[i].c &&
			    strncmp (cmds[i].word, cmd, strlen (cmd)) == 0)
				break;
		}

		sig_interrupted = 0;
		if (cmds[i].word != 0) {
			(*cmds[i].action) (&fd, addr, line, strlen (line));
		} else {
			(void) fprintf (stderr,
			    "? %s - not a command (quit to quit, ? for help)\n",
				tenex?"":cmd);
		}
	}
}

/* ARGSUSED */
static void nullcmd (fd, addr, line, len)
int *fd;
char *line;
int len;
{
}

/* ARGSUSED */
static void quitcmd (fd, addr, line, len)
int *fd;
char *line;
int len;
{
	quit (0);
}

/* ARGSUSED */
static void helpcmd (fd, addr, line, len)
int *fd;
char *line;
int len;
{
	int i;
	struct cmdtable *cmds = topcmds;
	int indent = 20;

	(void) printf ("Commands:\n");
	for (i = 0; cmds[i].word != 0; i++) {
		if (cmds[i].help != 0) {
			(void) printf ("%s %-*s %s\n",
				cmds[i].word,
				indent-strlen (cmds[i].word),
				cmds[i].msgset?"msgset":"",
				cmds[i].help);
		}
	}
	(void) printf ("\nMessage sets:\n");
	for (i = 0; ops[i].op != 0; i++) {
		if (ops[i].help != 0) {
			(void) printf ("%s %-*s %s\n",
				ops[i].op,
				indent-strlen (ops[i].op),
				(ops[i].expectarg==1)?"(string)":
				    (ops[i].expectarg==2)?"(msgset)":"",
				ops[i].help);
		}
	}
	(void) printf ("\nSet expressions:\n");
	(void) printf ("| %-*s or,  set union         (also ||)\n", indent, "");
	(void) printf ("& %-*s and, set intersection  (also &&)\n", indent, "");
	(void) printf ("! %-*s not\n", indent, "");
	(void) printf ("()%-*s parentheses for grouping\n", indent, "");
}

/* ARGSUSED */
static void typecmd (fd, addr, line, len)
int *fd;
struct sockaddr_in *addr;
char *line;
int len;
{
	char *nitype = getenv ("NITYPE");
	char *nline = malloc ((unsigned) strlen (line) + 7);
	register char *p = line;
	register char *q = nline;
	int pid, dead;
	int pipefds[2];
	union wait status;

	if (nline == 0) {
		perror ("?ni: malloc");
		quit (2);
	}

	(void) strcpy (q, "files ");
	while (isspace (*p))
		p++;
	while (isalnum (*p))
		p++;
	(void) strcat (q, p);

	if (pipe (pipefds) < 0) {
		perror ("?ni: pipe");
		quit (2);
	}

	if (nitype == 0)
		pid = fork ();
	else
		pid = vfork ();

	if (pid < 0) {
		perror ("?ni:fork");
		free (nline);
		return;
	}

	if (pid == 0) {
		(void) close (pipefds[1]);
		if (pipefds[0] != 0) {
			(void) dup2 (pipefds[0], 0);
			(void) close (pipefds[0]);
		}
		if (nitype == 0) {
			xargs ();
			(void) _exit (2);
		} else {
			(void) execlp (nitype, nitype, (char *) 0);
			(void) write (2, "?", 1);
			perror (nitype);
			(void) _exit (2);
		}
	}

	(void) close (pipefds[0]);

	(void) signal (SIGPIPE, sigint);
	send (fd, addr, nline, strlen (nline), pipefds[1]);
	(void) signal (SIGPIPE, SIG_DFL);

	(void) close (pipefds[1]);
	free (nline);

	do {
		dead = wait (&status);
	} while (dead != pid && (dead > 0 || errno == EINTR));
}

#define MAXARGS 50

static void xargs ()
{
	char *argv[MAXARGS+3];
	char buf[4096];
	char *p = buf;
	int argn = 2;
	int len;
	char *pager = getenv ("PAGER");

	if (pager == 0)
		pager = "more";

	if (chdir ("/usr/spool/news") < 0) {
		perror ("?: ni: can't chdir to /usr/spool/news");
		exit (2);
	}

	(void) signal (SIGTSTP, SIG_DFL);
	(void) signal (SIGINT, SIG_DFL);

	while (fgets (p, sizeof (buf) - (p - buf), stdin) != 0) {
		len = strlen (p);
		if (p[len-1] == '\n')
			p[len-1] = '\0';
		argv[argn++] = p;
		p += len+1;
		if (argn == MAXARGS || sizeof (buf) - (p-buf) < 1024) {
			argv[argn] = 0;
			argv[1] = pager;
			page (&argv[1]);
			argn = 2;
			p = buf;
		}
	}
	if (argn > 2) {
		(void) close (0);
		(void) open ("/dev/tty", 0);
		argv[argn] = 0;
		argv[1] = pager;
		(void) execvp (argv[1], &argv[1]);
		(void) write (2, "?", 1);
		perror (argv[1]);
		(void) _exit (2);
	}
	(void) _exit (0);
}

static void page (argv)
char *argv[];
{
	int pid;
	int dead;
	union wait status;
	
	if ((pid = vfork ()) < 0) {
		perror ("?ni: vfork");
		exit (2);
	}

	if (pid == 0) {
		(void) close (0);
		(void) open ("/dev/tty", 0);
		(void) execvp (argv[0], argv);
		(void) write (2, "?", 1);
		perror (argv[0]);
		exit (2);
	}

	do {
		dead = wait (&status);
	} while (dead != pid && (dead > 0 || errno == EINTR));
}


static void sendcmd (fd, addr, line, len)
int *fd;
struct sockaddr_in *addr;
char *line;
int len;
{
	send (fd, addr, line, len, 1);
}

static void send (fd, addr, line, len, outfd)
int *fd;
struct sockaddr_in *addr;
char *line;
int len;
int outfd;
{
	int retry;
	char buf[8192];
	int nread;
	int which;
	int end;
	int done;
	int wrote;
	int totty = isatty (outfd);
	int towrite;

	for (retry = 0, end = 0; !end && retry < 3; retry++) {
		if (*fd < 0)
			*fd = makesocket (addr);
		done = 0;
		while (done < len) {
			if (sig_interrupted) {
				(void) close (*fd);
				*fd = -1;
				return;
			}
			wrote = intwrite (*fd, line+done, len-done);
			if (wrote > 0)
				done += wrote;
			else if (errno != EINTR) {
				(void) close (*fd);
				*fd = -1;
				break;
			}
		}
		if (*fd < 0)
			continue;
		if (sig_interrupted) {
			(void) close (*fd);
			*fd = -1;
			return;
		}

		nread = 0;
		which = -1;
		while (!end &&
		       (nread = intread (*fd, buf, sizeof (buf))) > 0 ||
		       (nread < 0 && errno == EINTR)) {
			if (sig_interrupted) {
				(void) close (*fd);
				*fd = -1;
				return;
			}
			if (nread < 0)
				continue;
			if (buf[nread-1] == '\0') {
				nread--;
				end = 1;
			}

			if (which < 0 && nread > 0)
				which = (buf[0]=='?') ? 2: outfd;

			done = 0;
			while (done < nread) {
				if (sig_interrupted) {
					(void) close (*fd);
					*fd = -1;
					return;
				}
				towrite = nread-done;
				if (totty && towrite > 128)
					towrite = 128;
				wrote = intwrite (which, buf+done, towrite);
				if (wrote > 0)
					done += wrote;
				else if (errno != EINTR) {
					perror ("?ni: write");
					quit (2);
				} else if (which == outfd && totty) {
					int x = FWRITE;
					(void) ioctl (1, TIOCFLUSH, (char *)&x);
				}
			}
		}
		if (nread <= 0) {
			(void) close (*fd);
			*fd = -2;
		}
	}

	if (!end) {
		perror (*fd == -1? "?ni: write" : "?ni: read");
		quit (2);
	}
}


static char *fgetline (fp)
register FILE *fp;
{
	register int len = 0;
	unsigned max = 0;
	register char *p = 0;
	register int c;

	while ((c = getc (fp)) != EOF) {
		if (len+2 >= max) {
			max += 1024;
			p = realloc (p, max);
			if (p == 0) {
				perror ("?ni: realloc");
				quit (2);
			}
		}
		p[len++] = c;
		if (c == '\n') {
			p[len] = '\0';
			return (p);
		}
	}

	if (len > 0) {
		if (len+2 >= max) {
			max += 1024;
			p = realloc (p, max);
			if (p == 0) {
				perror ("?ni: realloc");
				quit (2);
			}
		}
		p[len++] = '\n';
		p[len] = '\0';
	} 

	return (p);
}

static char *agetline (args)
struct args *args;
{
	register int len = 0;
	unsigned max = 0;
	register char *p = 0;
	register int argn = args->argn;
	register int i = args->i;
	register int c;
	register char **argv = args->argv;
	int argc = args->argc;

	while (argn < argc) {
		while ((c = argv[argn][i++]) != '\0') {
			if (len+2 >= max) {
				max += 1024;
				p = realloc (p, max);
				if (p == 0) {
					perror ("?ni: realloc");
					quit (2);
				}
			}
			p[len++] = c;
			if (c == '\n') {
				if (len == 1)
					len = 0;
				else {
					p[len] = '\0';
					args->argn = argn;
					args->i = i;
					return (p);
				}
			}
		}
		i = 0;
		argn++;
	}

	args->argn = argn;
	args->i = i;

	if (len > 0) {
		if (len+2 >= max) {
			max += 1024;
			p = realloc (p, max);
			if (p == 0) {
				perror ("?ni: realloc");
				quit (2);
			}
		}
		p[len++] = '\n';
		p[len] = '\0';
	} 

	return (p);
}


static struct sgttyb ttyb, ottyb;
static struct tchars tch, otch;
static struct ltchars ltch = {-1, -1, -1, -1, -1, -1}, oltch;
static int ldisc, lmode;
static int hasttyinfo = 0;

static int gettty (fd, outfd)
register int fd, outfd;
{
	char *term = getenv ("TERM");

	if (term == 0)
		term = "";

	if (ioctl (outfd, (int) TIOCGETP, (char *) &ottyb) == -1)
		return (-1);
	if (ioctl (fd, (int) TIOCGETP, (char *) &ottyb) == -1)
		return (-1);
	(void) ioctl (fd, (int) TIOCGETC, (char *) &otch);
	(void) ioctl (fd, (int) TIOCGLTC, (char *) &oltch);
	(void) ioctl (fd, (int) TIOCGETD, (char *) &ldisc);
	(void) ioctl (fd, (int) TIOCLGET, (char *) &lmode);
	ttyb = ottyb;
	ttyb.sg_flags |= CBREAK;
	ttyb.sg_flags &= ~(RAW|ECHO);
	if ((ttyb.sg_flags & TBDELAY) == XTABS)
		ttyb.sg_flags &= ~TBDELAY;
	tch = otch;
	/* tch.t_intrc = -1; */
	/* tch.t_quitc = -1; */
	ltch.t_suspc = oltch.t_suspc;
	if ((erasechar = ottyb.sg_erase) == 0)
		erasechar = '\177';
	hasttyinfo = 1;
	interactive = 1;
	if (tenex < 0 && ldisc == NTTYDISC &&
	    (ottyb.sg_flags & ECHO) && (lmode & LCRTBS) && (lmode & LCRTERA) &&
	    strcmp (term, "ts") != 0 && strcmp (term, "typescript") != 0) {
		if (ioctl (fd, (int) TIOCSETN, (char *) &ttyb) == 0)
			tenex = 1;
	}
	return (0);
}

static void fixtty (fd, new)
register int fd, new;
{
	if (hasttyinfo) {
		(void) ioctl (fd, (int) TIOCSETN, (char *) (new?&ttyb:&ottyb));
		(void) ioctl (fd, (int) TIOCSETC, (char *) (new? &tch: &otch));
		(void) ioctl (fd, (int) TIOCSLTC, (char *) (new? &ltch:&oltch));
	}
}

#define MAXHIST	10
#define MAXCON	256

static int tenexmsgset (ifp, ofp, buf, len)	
FILE *ifp;
FILE *ofp;
char *buf;
int len;
{
	int c = 0;
	int i, j;
	int pos = 0;
	int end = 0;
	int level = 0;
	enum states { top, aftertop, input, variable } state = top;
	struct context {
		enum states state;
		int pos;
		int end;
		int level;
	} context[MAXCON];
	int con = 0;
	char *text;
	static char *brks[] = {
	"\1@'!(abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ\177\033 \t",
	"\1@'!&|()abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ\177\033 \t",
	"\1\177",
	"\1\177"
	};
	char tmpbuf[256];
	static struct hist {
		int con;
		struct context *context;
		char *buf;
	} history[MAXHIST];
	int histpos = 0;
	int prev = 0, prevcon = con;

	for (i = 0; i < 4; i++)
		brks[i][0] = erasechar;

	while (c != '\n') {
		c = inputstring (ifp, ofp, buf+pos, end, end,
			len-pos-level-10, "\n\t)", brks[(int) state], "\\");

		text = 0;
		context[con].state = state;
		context[con].pos = pos;
		context[con].end = strlen (buf+pos);
		context[con].level = level;

		switch (c)
		{
		case INTR:
		case EOF:
			return (c);
		case '\033':
			if (prev == c && prevcon < con) {
				con = prevcon;
				while (pos > context[con].pos + context[con].end) {
					(void) putc ('\b', ofp);
					(void) putc (' ', ofp);
					(void) putc ('\b', ofp);
					pos--;
				}
			} else if (state == aftertop) {
				text = " && ";
				state = top;
				c = '&';
				break;
			} else
				histpos = 0;
			if (histpos >= MAXHIST || history[histpos].buf == 0) {
				(void) putc ('\007', ofp);
				histpos = 0;
				break;
			}
			if (con + history[histpos].con+1 >= MAXCON ||
			    strlen (history[histpos].buf)+pos+level+10 >= len) {
				(void) putc ('\007', ofp);
				break;
			}
			(void) strcpy (buf+pos, history[histpos].buf);
			(void) fprintf (ofp, "%s", buf+pos);
			bcopy ((char *) history[histpos].context,
				(char *) &context[con],
				history[histpos].con * sizeof (struct context));
			prevcon = con;
			for (i = 0; i < history[histpos].con; i++) {
				context[con].pos += pos;
				context[con].level += level;
				con++;
			}
			con--;
			histpos++;
			break;
		case '\177':
		delete_char:
			if (con > 0) {
				con--;
				while (pos > context[con].pos + context[con].end) {
					(void) putc ('\b', ofp);
					(void) putc (' ', ofp);
					(void) putc ('\b', ofp);
					pos--;
				}
			} else
				return (c);
			break;
		case '!':
			if (state == aftertop) {
				text = " && ! ";
				state = top;
			} else
				text = "! ";
			break;
		case '(':
			if (state == aftertop) {
				text = " && (";
				state = top;
			} else
				text = "(";
			level++;
			break;
		case ')':
			if (state == input) {
				(void) strcat (buf+pos, ")");
				(void) putc (')', ofp);
				text = buf+pos;
				state = aftertop;
			} else if (level > 0) {
				text = ")";
				level--;
			}
			break;
		case '\n':
		case '\t':
			if (c == '\n') {
				while (level > 0) {
					(void) strcat (buf+pos, ")");
					(void) putc (')', ofp);
					level--;
				}
			}
			if (state == input) {
				(void) strcat (buf+pos, ")");
				(void) putc (')', ofp);
			}
			if (c == '\n' || state == input) {
				text = buf+pos;
				state = aftertop;
			}
			break;
		case '&':
			text = " && ";
			state = top;
			break;
		case '|':
			text = " || ";
			state = top;
			break;
		case '\'':
		case '@':
			tmpbuf[0] = '\0';
			if (state == aftertop)
				(void) strcat (tmpbuf, " && ");
			(void) strcat (tmpbuf, c=='@'?"@":"'");
			state = variable;
			text = tmpbuf;
			break;

		default:
			if (c == erasechar)
				goto delete_char;
			tmpbuf[0] = '\0';
			if (state == aftertop) {
				(void) strcat (tmpbuf, " && ");
				state = top;
			}
			for (i = 0; ops[i].op != 0; i++)
				if (ops[i].tc == c)
					break;
			if (ops[i].op != 0) {
				(void) strcat (tmpbuf, ops[i].op);
				if (ops[i].expectarg == 1) {
					(void) strcat (tmpbuf, " (");
					state = input;
				} else if (ops[i].expectarg == 2) {
					(void) strcat (tmpbuf, " (");
					level++;
					state = top;
				} else if (level == 0) {
					c = '\n';
					state = aftertop;
				} else {
					state = aftertop;
				}
				text = tmpbuf;
			}
			break;
		}

		if (text != 0 && pos + level + strlen (text) + 2 <= len) {
			con++;
			if (text != buf+pos) {
				(void) fprintf (ofp, "%s", text);
				(void) strcpy (buf+pos, text);
			}
			pos += strlen (text);
			end = 0;
		} else {
			state = context[con].state;
			pos = context[con].pos;
			level = context[con].level;
			end = context[con].end;
		}
		prev = c;
	}

	if (strlen (buf) > 0) {
		if (con == 0 || context[con-1].pos != pos) {
			context[con].state = state;
			context[con].pos = pos;
			context[con].level = level;
			context[con].end = end;
			con++;
		}

		for (j = 0; j < MAXHIST; j++) {
			if (history[j].buf != 0)
				if (strcmp (buf, history[j].buf) == 0)
					break;
		}
		if (j >= MAXHIST)
			j = MAXHIST-1;

		if (history[j].buf != 0)
			free (history[j].buf);
		if (history[j].context != 0)
			free ((char *) history[j].context);

		for (i = j; i > 0; i--)
			history[i] = history[i-1];

		history[0].con = con;
		history[0].context = (struct context *)
			malloc ((unsigned) con * sizeof (struct context));
		history[0].buf = malloc ((unsigned) strlen (buf) + 1);
		if (history[0].context == 0 || history[0].buf == 0) {
			perror ("?ni: malloc");
			quit (2);
		}
		(void) strcpy (history[0].buf, buf);
		bcopy ((char *) context, (char *) history[0].context,
			con * sizeof (struct context));
	}

	return (c);
}

static int isescape (c, string, pos)
register char  c;
register char *string;
register int pos;
{
	register int i;

	if (c == 0)
		return (0);

	for (i = 0; i < pos && string[i] != '\0'; i++) {
		if (string[i] == c)
			i++;
	}
	return (i == pos && string[i] == c);
}

static int inputstring (ifp, ofp, buf, pos, end, maxlen, brk, ibrk, esc)
FILE *ifp;
FILE *ofp;
char *buf;
int pos;
int end;
int maxlen;
char *brk;
char *ibrk;
char *esc;
{
	int i;
	int c;

	for (;;) {
		(void) fflush (ofp);
		c = intgetc (ifp);

		if (c == EOF || c == INTR ||
		   (index (brk, c) != 0 &&
		      (pos == 0 || !isescape (esc[0], buf, pos-1))) || 
		   (end == 0 && pos == 0 && index (ibrk, c) != 0)) {
			if (c != INTR)
				for (i = pos; i < end; i++)
					(void) putc (buf[i], ofp);
			buf[end] = '\0';
			return (c);
		}

		if (!isascii (c) || c == '\0')
			;
		else if (c == '\177' || c == erasechar) {
			if (pos > 0) {
				int sub = (pos > 1 &&
					isescape (esc[0], buf, pos-2))? 2 : 1;

				pos -= sub;
				for (i = 0; i < sub; i++)
					(void) putc ('\b', ofp);
				for (i = pos+sub; i < end; i++) {
					buf[i-sub] = buf[i];
					(void) putc (buf[i], ofp);
				}
				for (i = 0; i < sub; i++)
					(void) putc (' ', ofp);
				for (i = pos; i < end; i++)
					(void) putc ('\b', ofp);
				end -= sub;
			}
		} else if (c == '\b' || c == '\002') {
			if (pos > 0) {
				if (pos == end && buf[pos-1] == ' ')
					end--;
				pos--;
				(void) putc ('\b', ofp);
				if (pos > 0 && isescape (esc[0], buf, pos-1)) {
					pos--;
					(void) putc ('\b', ofp);
				}
			}
		} else if (c == '\014' || c == '\006') {
			if (pos < end) {
				(void) putc (buf[pos], ofp);
				pos++;
				if (pos < end && isescape (esc[0], buf, pos-1)) {
					(void) putc (buf[pos], ofp);
					pos++;
				}
			} else if (end < maxlen-1) {
				(void) putc (' ', ofp);
				buf[pos] = ' ';
				pos++, end++;
			}
		} else if (c == ' ' || (isprint (c) && !isspace (c))) {
			if (end < maxlen-1) {
				(void) putc (c, ofp);
				for (i = pos; i < end; i++)
					(void) putc (buf[i], ofp);
				for (i = end-1; i >= pos; i--) {
					buf[i+1] = buf[i];
					(void) putc ('\b', ofp);
				}
				buf[pos] = c;
				pos++, end++;
			}
		}
	}
}

static char *tgetline (fp)
register FILE *fp;
{
	char *cmdbuf = malloc (1024);

	if (cmdbuf == 0) {
		perror ("?ni: malloc");
		quit (2);
	}

	dotenex (fp, stderr, cmdbuf, 1024, topcmds);

	return (cmdbuf);
}

static void dotenex (ifp, ofp, buf, len, cmds)
FILE *ifp;
FILE *ofp;
char *buf;
int len;
struct cmdtable *cmds;
{
	int c;
	int i;
	int pos = 0;
	int end = 0;
	enum states { top, input, inputmsgset, done } state = top;
	struct context {
		enum states state;
		int pos;
		int end;
	} context[20];
	int con = 0;
	char *text, *textextra;
	static char erasestr[] = { '\177', '\177', '\0' };

	erasestr[0] = erasechar;

	for (;;) {
		text = 0;
		textextra = "";
		context[con].state = state;
		context[con].pos = pos;
		context[con].end = 0;

		switch (state)
		{
		case top:
			(void) fflush (ofp);
			c = intgetc (ifp);
			buf[pos] = '\0';
			break;
		case input:
			c = inputstring (ifp, ofp, buf+pos, end, end, 
				    len-pos-10, "\n\t\"", erasestr, "\\");
			context[con].end = strlen (buf+pos);
			break;
		case inputmsgset:
			c = tenexmsgset (ifp, ofp, buf+pos, len-pos-1);
			context[con].end = strlen (buf+pos);
			break;
		case done:
			buf[pos++] = '\n';
			buf[pos] = '\0';
			(void) fprintf (ofp, "\n");
			(void) fflush (ofp);
			return;
		default:
			fixtty (fileno (ifp), 0);
			abort ();
		}

		if (c == INTR) {
			(void) putc ('\n', stderr);
			buf[0] = '\0';
			return;
		} else if (c == '\177' || c == erasechar) {
			if (con > 0) {
				con--;
				while (pos> context[con].pos+context[con].end) {
					(void) putc ('\b', ofp);
					(void) putc (' ', ofp);
					(void) putc ('\b', ofp);
					pos--;
				}
			} 
		} else if (state == top) {
			for (i = 0; cmds[i].word != 0 && c != cmds[i].c; i++)
				;

			if (cmds[i].word == 0) {
				(void) strcpy (buf, chartos (c));
				(void) fprintf (ofp, "%s", buf);
				return;
			}
			text = cmds[i].word;
			if (cmds[i].arg) {
				state = input;
				textextra = " \"";
			} else if (cmds[i].msgset) {
				state = inputmsgset;
				textextra = " ";
			} else {
				state = done;
				textextra = "";
			}
		} else if (state == input) {
			text = buf+pos;
			if (cmds[i].msgset) {
				state = inputmsgset;
				textextra = "\" ";
			} else {
				state = done;
				textextra = "";
			}
		} else if (state == inputmsgset) {
			text = buf+pos;
			textextra = "";
			state = done;
		}

		if (text != 0 && pos+strlen (text)+strlen (textextra)+2 < len) {
			con++;
			if (text != buf+pos) {
				(void) fprintf (ofp, "%s", text);
				(void) strcat (buf+pos, text);
			}
			(void) fprintf (ofp, "%s", textextra);
			(void) strcat (buf+pos, textextra);
			pos += strlen (buf+pos);
			end = 0;
		} else {
			state = context[con].state;
			pos = context[con].pos;
			end = context[con].end;
		}
	}
}

static char *getarg (line, buf, size)
register char *line;
register char *buf;
int size;
{
	static char quotes[] = "\"'(";
	static char endquotes[] = "\"')";
	char *p;
	register int endquote;
	register int i = 0;

	size--;
	while (*line != '\0' && isspace (*line))
		line++;

	if (*line != '\0') {
		p = index (quotes, *line);

		if (p == 0)
			endquote = ' ';
		else {
			line++;
			endquote = endquotes[p-quotes];
		}

		while (*line != '\0' && *line != endquote &&
			(endquote != ' ' || isalnum (*line) || *line == '?')) {
			if (*line == '\\' && line[1] != '\0')
				line++;
			if (i < size)
				buf[i++] = *line;
			line++;
		}

		if (*line == endquote)
			line++;
		while (*line != '\0' && isspace (*line))
			line++;
	}
	buf[i] = '\0';

	return (line);
}

static char *chartos (c)
int c;
{
	static char buf[10];

	if (c == EOF)
		return ("EOF");

	if (isascii (c) && isprint (c) && !isspace (c)) {
		buf[0] = c;
		buf[1] = '\0';
		return (buf);
	}
	if (c == ' ')
		return ("SPACE");
	if (c == '\t')
		return ("TAB");
	if (c == '\n')
		return ("NL");
	if (c == '\r')
		return ("CR");
	(void) sprintf (buf, "\\%03o", c);
	return (buf);
}

static int sig_returnset = 0;
static jmp_buf sig_return;

/* ARGSUSED */
static void sigint (sig)
int sig;
{
	if (sig_nostop) {
		sig_interrupt = 1;
		return;
	}
	sig_interrupt = 0;
	sig_interrupted = 1;

	if (!interactive)
		quit (2);

	if (sig_returnset) {
		sig_returnset = 0;
		longjmp (sig_return, 1);
	}
}

/* ARGSUSED */
static void sigtstp (sig)
int sig;
{
	int omask;

	if (sig_nostop) {
		sig_stop = 1;
		return;
	}
	sig_stop = 0;

	if (tenex)
		fixtty (0, 0);
	(void) signal (SIGTSTP, SIG_DFL);
	omask = sigsetmask (0);
	(void) kill (0, SIGTSTP);
	(void) sigsetmask (omask);
	(void) signal (SIGTSTP, sigtstp);
	if (tenex)
		fixtty (0, 1);

	if (sig_returnset) {
		sig_returnset = 0;
		longjmp (sig_return, 1);
	}
}

static void endregion ()
{
	sig_nostop = 0;
	if (sig_stop)
		sigtstp (SIGTSTP);
	if (sig_interrupt)
		sigint (SIGINT);
}

static int intread (fd, buf, count)
int fd;
char *buf;
int count;
{
	register int nread;

	if (setjmp (sig_return) == 0) {
		sig_returnset = 1;
		nread = read (fd, buf, count);
		sig_returnset = 0;
		return (nread);
	}
	errno = EINTR;
	return (-1);
}

static int intwrite (fd, buf, count)
int fd;
char *buf;
int count;
{
	register int nwrote;

	if (setjmp (sig_return) == 0) {
		sig_returnset = 1;
		nwrote = write (fd, buf, count);
		sig_returnset = 0;
		return (nwrote);
	}
	errno = EINTR;
	return (-1);
}

static int intgetc (fp)
FILE *fp;
{
	char c;
	int nread;

	nread = intread (fileno (fp), &c, 1);

	if (nread < 0 && errno == EINTR)
		return (INTR);
	if (nread <= 0)
		return (EOF);
	return (c);
}
